import sys
from ._types import Address, AddressRange, NativeNumber
from typing import Tuple, List

if sys.version_info[0] == 3 and sys.version_info[1] == 7:
    class Protocol:
        pass
else:
    from typing import Protocol


class Slave(Protocol):
    def __setitem__(self, address: Address, value: NativeNumber) -> None:
        raise NotImplementedError()

    def __getitem__(self, address: Address) -> NativeNumber:
        raise NotImplementedError()


class Bus:
    def __init__(self):
        self._attached: List[Tuple[AddressRange, Slave]] = []

    def attach(self, address_range: AddressRange, slave: Slave):
        self._attached.append((address_range, slave))

    def __setitem__(self, address: Address, value: NativeNumber):
        for address_range, slave in self._attached:
            if address in address_range:
                slave[Address(address.value - address_range.start_value)] = value
                return
        raise ValueError('Invalid address')

    def __getitem__(self, address: Address):
        for address_range, slave in self._attached:
            if address in address_range:
                return slave[Address(address.value - address_range.start_value)]
        raise ValueError('Invalid address')



__version__ = '0.0.1'

from .cpu import CPU, Instructions
from .bus import Bus
from .ram import RAM
from .vm import VM
from ._types import Address, NativeNumber, AddressRange, NativeFalse, NativeTrue



import time
import sys
from ._types import NativeNumber, Address, AddressRange
from .cpu import CPU, HaltExecution
from .ram import RAM
from .bus import Bus
from itertools import count
from enum import Enum


class VM:
    def __init__(self):
        self._fsb = Bus()
        self._ram = RAM(256)
        self._fsb.attach(AddressRange(0, 256), self._ram)
        self._cpu = CPU(self._fsb)

    def run(self, frequency=None):
        try:
            if frequency is None:
                while True:
                    self._cpu.next_instruction()
            else:
                period_ns = int(1000000000.0 / frequency)
                while True:
                    cycle_start_ts_ns = time.perf_counter_ns()
                    self._cpu.next_instruction()
                    cycle_overtime_ns = period_ns - (time.perf_counter_ns() - cycle_start_ts_ns)
                    if cycle_overtime_ns >= 0:
                        time.sleep(cycle_overtime_ns * 0.000000001)
                    else:
                        print(self._cpu, 'throttling to', 1000000000.0 / (period_ns - cycle_overtime_ns), 'Hz',
                              file=sys.stderr)
        except HaltExecution:
            pass

    def reset(self):
        self._ram.clear()
        self._cpu.reset()

    def load_program(self, program):
        assert len(program) <= len(self._ram)
        for address, value in zip(count(), program):
            if isinstance(value, Enum):
                value = value.value
            self._ram[Address(address)] = NativeNumber(value)

    def __getitem__(self, item: Address) -> NativeNumber:
        return self._fsb[item]

    def __repr__(self):
        return '\n\n'.join([self._cpu.__repr__(), self._ram.__repr__()])



from .bus import Bus
from ._types import Address, NativeNumber, NativeFalse, NativeTrue, float_to_native_number
from enum import Enum, auto
from typing import Dict, Callable, Tuple, Optional
from inspect import getfullargspec
from math import sqrt


class Instructions(Enum):
    Halt = auto()

    St = auto()  # store Acc value to Address(a)
    Ld = auto()  # load Address(a) value to Acc

    Add = auto()  # Acc += a
    Neg = auto()  # Acc = -Acc
    Mul = auto()  # Acc *= a
    Div = auto()  # Acc //= a
    Sqrt = auto()  # sqrt(Acc)

    Eq = auto()  # Acc = 1 if Acc == a else 0
    Gt = auto()  # Acc = 1 if Acc > a else 0

    Not = auto()  # Acc = 1 if Acc == 0 else 0
    Or = auto()  # Acc = 1 if Acc or a else 0
    And = auto()  # Acc = 1 if Acc and a else 0

    Jmp = auto()  # IA = a
    Jif = auto()  # if Acc != 0: IA = a


_instruction_methods: Dict[Instructions, Tuple[Callable, int]] = {}


def perform_instruction(name: Instructions):
    def decorator(method: Callable):
        argspec = getfullargspec(method)
        assert name not in _instruction_methods, 'Instruction redefinition'
        _instruction_methods[name] = (method, len(argspec.args) - 1)
        return method

    return decorator


class InvalidInstruction(Exception):
    def __init__(self, address: Address, opcode: Optional[int]):
        self.address = address
        self.opcode = opcode


class HaltExecution(Exception):
    pass


class CPU:
    def __init__(self, bus: Bus):
        self._bus = bus
        self._IA = Address()  # next instruction address
        self._Acc = NativeNumber()  # accumulator
        self.reset()

    def reset(self):
        self._IA = Address(0)
        self._Acc = NativeNumber(0)

    def next_instruction(self):
        instruction_address = self._IA

        opcode = None
        try:
            opcode = self._bus[instruction_address]
            instruction = Instructions(opcode.value)
        except ValueError:
            raise InvalidInstruction(instruction_address, opcode)

        method, args_num = _instruction_methods[instruction]
        args = list(map(lambda a: Address(int(self._bus[Address(a)].value)),
                        range(instruction_address.value + 1,
                              instruction_address.value + 1 + args_num)))

        self._IA = Address(self._IA.value + 1 + args_num)
        method(self, *args)

    @perform_instruction(Instructions.Halt)
    def _halt(self):
        raise HaltExecution()

    @perform_instruction(Instructions.St)
    def _store(self, address: Address):
        self._bus[address] = self._Acc

    @perform_instruction(Instructions.Ld)
    def _load(self, address: Address):
        self._Acc = self._bus[address]

    @perform_instruction(Instructions.Add)
    def _add(self, address: Address):
        self._Acc = NativeNumber(self._Acc.value + self._bus[address].value)

    @perform_instruction(Instructions.Neg)
    def _neg(self):
        self._Acc = NativeNumber(-self._Acc.value)

    @perform_instruction(Instructions.Mul)
    def _multiply(self, address: Address):
        self._Acc = NativeNumber(self._Acc.value * self._bus[address].value)

    @perform_instruction(Instructions.Div)
    def _divide(self, address: Address):
        self._Acc = float_to_native_number(self._Acc.value / self._bus[address].value)

    @perform_instruction(Instructions.Sqrt)
    def _square_root(self):
        self._Acc = float_to_native_number(sqrt(self._Acc.value))

    @perform_instruction(Instructions.Eq)
    def _equal(self, address: Address):
        self._Acc = NativeTrue if self._Acc == self._bus[address] else NativeFalse

    @perform_instruction(Instructions.Gt)
    def _greater(self, address: Address):
        self._Acc = NativeTrue if self._Acc.value > self._bus[address].value else NativeFalse

    @perform_instruction(Instructions.Not)
    def _not(self):
        self._Acc = NativeTrue if self._Acc.value == NativeFalse.value else NativeFalse

    @perform_instruction(Instructions.Or)
    def _or(self, address: Address):
        self._Acc = NativeTrue if self._Acc.value or self._bus[address].value else NativeFalse

    @perform_instruction(Instructions.And)
    def _and(self, address: Address):
        self._Acc = NativeTrue if self._Acc.value and self._bus[address].value else NativeFalse

    @perform_instruction(Instructions.Jmp)
    def _jump(self, address: Address):
        self._IA = address

    @perform_instruction(Instructions.Jif)
    def _jump_if(self, address: Address):
        if self._Acc:
            self._IA = address

    def __dict__(self):
        return {
            'IA': self._IA.value,
            'Acc': self._Acc.value,
        }

    def __str__(self):
        return 'CPU(' + ', '.join(map(lambda item: f'{item[0]}: {item[1]:02x}', self.__dict__().items())) + ')'

    def __repr__(self):
        return 'CPU:\n' + '\n'.join(map(lambda item: f'    {item[0]}: {item[1]:02x}', self.__dict__().items()))



from ._types import Address, NativeNumber
from .bus import Slave
from itertools import count
from ctypes import memset, sizeof


class RAM(Slave):
    def __init__(self, capacity: int):
        self._capacity = capacity
        self._cells = (NativeNumber * capacity)()
        self.clear()

    def __getitem__(self, address: Address) -> NativeNumber:
        return NativeNumber(self._cells[address.value])

    def __setitem__(self, address: Address, value: NativeNumber) -> None:
        self._cells[address.value] = value

    def clear(self):
        memset(self._cells, 0, self._capacity * sizeof(NativeNumber))

    def __len__(self):
        return self._capacity

    def __repr__(self):
        line_segment_len = 2
        line_segments_num = 16
        # noinspection PyTypeChecker
        hex_str = bytes(self._cells).hex()
        header = list(map(lambda i: f'{i:02x}', range(line_segments_num))) + ['..'] * line_segments_num
        hex_str = header + [hex_str[i:i + line_segment_len]
                            for i in range(0, len(hex_str), line_segment_len)]
        hex_str = [' '.join(hex_str[i:i + line_segments_num])
                   for i in range(0, len(hex_str), line_segments_num)]
        hex_str = '\n'.join(map(lambda t: (
                                f'    {t[0] * line_segments_num:06x} : '
                                if t[0] >= 0 else
                                '             ')
                                + t[1],
                                zip(count(-2), hex_str)))
        return f'RAM({self._capacity}x{sizeof(NativeNumber)} bytes)\n{hex_str}'



import ctypes
from typing import Union

NativeNumber = ctypes.c_byte


def float_to_native_number(f):
    return NativeNumber(int(f))


def int_to_native_number(i):
    return NativeNumber(i)

# NativeNumber = ctypes.c_double
#
#
# def float_to_native_number(f):
#     return NativeNumber(f)
#
#
# def int_to_native_number(i):
#     return NativeNumber(float(i))


Address = ctypes.c_ubyte

NativeFalse = NativeNumber(0)
NativeTrue = NativeNumber(1)


class AddressRange:
    def __init__(self, start: Union[int, Address], end: Union[int, Address]):
        self.start_value = start.value if isinstance(start, Address) else start
        self.end_value = end.value if isinstance(end, Address) else end
        assert self.start_value <= self.end_value, 'Invalid range'

    def __contains__(self, item: Address):
        return self.start_value <= item.value < self.end_value



import setuptools

setuptools.setup(
    name="crash-vm",
    version="0.0.1",
    description="Simple virtual machine implemented in pure python",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages=["crash_vm"],
    python_requires='>=3.7',
)
